#!/usr/local/ymir/perl/bin/perl

use strict;

my $re_char = qr/[\x81-\x9f\xe0-\xef\xfa-\xfc][\x40-\x7e\x80-\xfc]|[\x00-\x7f]/;


our %HIRAKATA = (
 '' => '@',
 '' => 'A',
 '' => 'B',
 '' => 'C',
 '' => 'D',
 '' => 'E',
 '' => 'F',
 '' => 'G',
 '' => 'H',
 '' => 'I',
 '' => 'J',
 '' => 'K',
 '' => 'L',
 '' => 'M',
 '' => 'N',
 '' => 'O',
 '' => 'P',
 '' => 'Q',
 '' => 'R',
 '' => 'S',
 '' => 'T',
 '' => 'U',
 '' => 'V',
 '' => 'W',
 '' => 'X',
 '' => 'Y',
 '' => 'Z',
 '' => '[',
 '' => '\\',
 '' => ']',
 '' => '^',
 '' => '_',
 '' => '`',
 '' => 'a',
 '' => 'b',
 '' => 'c',
 '' => 'd',
 '' => 'e',
 '' => 'f',
 '' => 'g',
 '' => 'h',
 '' => 'i',
 '' => 'j',
 '' => 'k',
 '' => 'l',
 '' => 'm',
 '' => 'n',
 '' => 'o',
 '' => 'p',
 '' => 'q',
 '' => 'r',
 '' => 's',
 '' => 't',
 '' => 'u',
 '' => 'v',
 '' => 'w',
 '' => 'x',
 '' => 'y',
 '' => 'z',
 '' => '{',
 '' => '|',
 '' => '}',
 '' => '~',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
 '' => '',
);

our %TENSPLIT = (
 'K' => 'JJ',
 'M' => 'LJ',
 'O' => 'NJ',
 'Q' => 'PJ',
 'S' => 'RJ',
 'U' => 'TJ',
 'W' => 'VJ',
 'Y' => 'XJ',
 '[' => 'ZJ',
 ']' => '\J',
 '_' => '^J',
 'a' => '`J',
 'd' => 'cJ',
 'f' => 'eJ',
 'h' => 'gJ',
 'o' => 'nJ',
 'p' => 'nK',
 'r' => 'qJ',
 's' => 'qK',
 'u' => 'tJ',
 'v' => 'tK',
 'x' => 'wJ',
 'y' => 'wK',
 '{' => 'zJ',
 '|' => 'zK',
);



opendir(DIR, "hmo") or die;
my @files = sort grep(m/\.hmo_kana$/, readdir(DIR));
closedir(DIR);

my $testdata;
{
  local($/) = undef;
  $testdata = <>;
}

print "<html><body>\n";
print "<table border=1>\n";

my $keydump = {};
my @testst = split(m/[\n ]+/, $testdata);
my $teststmax = scalar(@testst);

my $linecount = 0;

foreach my $file (sort @files) {
  my ($conf, $table) = &loadkeytable("hmo/$file");
  print STDERR "==== $file\n";
  my $res = '';
  my $count = 0;
  my $oldper = 0;
  print STDERR "=" x 60, "\n";
  foreach my $st (@testst) {
    $count++;
    $res .= &parsekey($conf, $table, $st, \&keyscore);
    my $per = int(($count / $teststmax) * 60);
    if($per > $oldper) {
      print STDERR '.' x ($per - $oldper);
      $oldper = $per;
    }
  }
  print STDERR '.' x (60 - $oldper) if($oldper < 60);
  print STDERR "\n";
  $keydump->{$file} = $res;
  
  my $keycount = &keycount($res);
  my $movecount = &movecount($res);
  
  if($linecount++ % 10 == 0) {
    print "<tr bgcolor=\"#fff8ff\"><th>t@C</th><th>L[</th>";
    print "<th></th><th>E</th><th>Vtg</th><th></th><th></th><th></th><th>l</th><th>El</th><th>E</th><th>E</th><th>Eq</th></tr>";
    print "<tr bgcolor=\"#f8ffff\"><th></th><th>ړ</th><th>ŏ</th><th></th><th></th><th></th><th>ݑŌ</th><th></th><th>EE</th><th>A</th><th>wٌ</th><th>蒵<br>(2i)</th><th>蒵<br>(3i)</th>";
    print "</tr>\n";
  }
  
  print "<tr align=\"right\" bgcolor=\"#fff8ff\">";
  print "<td align=\"left\">$file</td>";
  print "<td>" . $keycount . "</td>";
  my $fingerdata = &fingerkeycount($conf, $res, $keycount, $movecount);
  foreach my $data (@$fingerdata[0..10]) {
    my $color = &hue($data->[0] / ($data->[1] <= 0 ? 1 : $data->[1]));
    print "<td><font color=\"$color\">" . $data->[0] . "<br>" . sprintf('%3.1f%%', (100 * $data->[0] / ($data->[1] <= 0 ? 1 : $data->[1]))) . "</font></td>";
  }
  
  print "</tr>\n<tr align=\"right\" bgcolor=\"#f8ffff\">";
  print '<td>' . ($conf->{StrokeMode} ? 'Xg[N' : 'Ō' ) . "</td>";
  print "<td>" . $movecount . "</td>";
  foreach my $data (@$fingerdata[11..21]) {
    my $color = &hue($data->[0] / ($data->[1] <= 0 ? 1 : $data->[1]));
    print "<td><font color=\"$color\">" . $data->[0] . "<br>" . sprintf('%3.1f%%', (100 * $data->[0] / ($data->[1] <= 0 ? 1 : $data->[1]))) . "</font></td>";
  }
  print "</tr>\n";
  
}

print "</table>\n";

foreach my $file (sort @files) {
  print "<hr>";
  print "<h3>$file</h3>\n";
  print $keydump->{$file}, "<br>\n";
}


print "</body></html>\n";


sub keyscore {
  my $key = shift;
  
  my $count = &keycount($key);
  my $move = &movecount($key);
  
  ($count == 0) ? 0 : $count + $move / $count;
}

sub keycount {
  my $key = shift;
  
  $key =~ s/\([^\)]+\)//g;
  $key =~ s/\s//g;
  
  length($key);
}

sub movecount {
  my $key = shift;
  
  $key =~ s/\([^\)]+\)//g;
  $key =~ s/[^\s]+//g;
  
  length($key) + 1;
}

sub fingerkeycount {
  my $conf = shift;
  my $key = shift;
  my $keycount = shift;
  my $movecount = shift;
  my $res = [];
  
  $key =~ s/\([^\)]+\)//g;
  $key = ' ' . $key;
  
# "<th></th><th>E</th><th>Vtg</th>";
  
  my $key2;
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/12345qwertasdfgzxcvb//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/67890yuiophjkl;nm,\.\/\-\^\\\@\[:\]_//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/12345qwertasdfgzxcvb67890yuiophjkl;nm,\.\/\-\^\\\@\[:\]_ //c), $keycount]);
  
# "<th></th><th></th><th></th><th>l</th>";
  
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/1qaz//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/2wsx//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/3edc//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/4rfv5tgb//), $keycount]);
  
# "<th>El</th><th>E</th><th>E</th><th>Eq</th>";
  
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/6yhn7ujm//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/8ik,//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/9ol\.//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/0p;\/\-\^\\\@\[:\]_//), $keycount]);
  
# "<th>ŏ</th><th></th><th></th><th></th>";
  
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/1234567890\-\^\\//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/qwertyuiop\@\[//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/asdfghjkl;:\]//), $keycount]);
  $key2 = $key;
  push(@$res, [scalar($key2 =~ tr/zxcvbnm,\.\/_//), $keycount]);
  
# "<th>ݑŌ</th><th></th><th>EE</th><th>A</th>";
  
  my $count;
  my @tmp;
  $key2 = $key;
  if($conf->{StrokeMode}) {
    my $key3 = $key2;
    $key3 =~ tr/ //d;
    $count = scalar(@tmp = $key3 =~ m/[12345qwertasdfgzxcvb](?=[67890yuiophjkl;nm,\.\/\-\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key3 =~ m/[67890yuiophjkl;nm,\.\/\-\^\\\@\[:\]_](?=[12345qwertasdfgzxcvb])/g);
    push(@$res, [scalar($count), $keycount - 1]);
    $count = scalar(@tmp = $key3 =~ m/[12345qwertasdfgzxcvb](?=[12345qwertasdfgzxcvb])/g);
    push(@$res, [scalar($count), $keycount - 1]);
    $count = scalar(@tmp = $key3 =~ m/[67890yuiophjkl;nm,\.\/\-\^\\\@\[:\]_](?=[67890yuiophjkl;nm,\.\/\-\^\\\@\[:\]_])/g);
    push(@$res, [scalar($count), $keycount - 1]);
    $count = scalar(@tmp = $key3 =~ m/([^ A-Z])(?=\1)/g);
    push(@$res, [scalar($count), $keycount - 1]);
  } else {
    $count = scalar(@tmp = $key2 =~ m/(?: )[^ ]*[12345qwertasdfgzxcvb](?= [67890yuiophjkl;nm,\.\/\-\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[67890yuiophjkl;nm,\.\/\-\^\\\@\[:\]_](?= [12345qwertasdfgzxcvb])/g);
    push(@$res, [scalar($count), $movecount - 1]);
    $count = scalar(@tmp = $key2 =~ m/(?: )[^ ]*[12345qwertasdfgzxcvb][^ ]*(?= [^ ]*[12345qwertasdfgzxcvb])/g);
    push(@$res, [scalar($count), $movecount - 1]);
    $count = scalar(@tmp = $key2 =~ m/(?: )[^ ]*[67890yuiophjkl;nm,\.\/\-\^\\\@\[:\]_][^ ]*(?= [^ ]*[67890yuiophjkl;nm,\.\/\-\^\\\@\[:\]_])/g);
    push(@$res, [scalar($count), $movecount - 1]);
    $count = scalar(@tmp = $key2 =~ m/(?: )[^ ]*([^ A-Z])[^ ]*(?= [^ ]*\1)/g);
    push(@$res, [scalar($count), $movecount - 1]);
  }

  
# "<th>wٌ</th><th>蒵(2i)</th><th>蒵(3i)</th>;
  
  if($conf->{StrokeMode}) {
    my $key3 = $key2;
    $key3 =~ tr/ //d;
    $count = scalar(@tmp = $key3 =~ m/[1](?=[qaz])/g)
           + scalar(@tmp = $key3 =~ m/[2](?=[wsx])/g)
           + scalar(@tmp = $key3 =~ m/[3](?=[edc])/g)
           + scalar(@tmp = $key3 =~ m/[4](?=[rfv])/g)
           + scalar(@tmp = $key3 =~ m/[5](?=[tgb])/g)
           + scalar(@tmp = $key3 =~ m/[6](?=[yhn])/g)
           + scalar(@tmp = $key3 =~ m/[7](?=[ujm])/g)
           + scalar(@tmp = $key3 =~ m/[8](?=[ik,])/g)
           + scalar(@tmp = $key3 =~ m/[9](?=[ol\.])/g)
           + scalar(@tmp = $key3 =~ m/[0](?=[p;\/\-\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key3 =~ m/[q](?=[1az])/g)
           + scalar(@tmp = $key3 =~ m/[w](?=[2sx])/g)
           + scalar(@tmp = $key3 =~ m/[e](?=[3dc])/g)
           + scalar(@tmp = $key3 =~ m/[r](?=[4fv])/g)
           + scalar(@tmp = $key3 =~ m/[t](?=[5gb])/g)
           + scalar(@tmp = $key3 =~ m/[y](?=[6hn])/g)
           + scalar(@tmp = $key3 =~ m/[u](?=[7jm])/g)
           + scalar(@tmp = $key3 =~ m/[i](?=[8k,])/g)
           + scalar(@tmp = $key3 =~ m/[o](?=[9l\.])/g)
           + scalar(@tmp = $key3 =~ m/[p](?=[0;\/\-\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key3 =~ m/[a](?=[1qz])/g)
           + scalar(@tmp = $key3 =~ m/[s](?=[2wx])/g)
           + scalar(@tmp = $key3 =~ m/[d](?=[3ec])/g)
           + scalar(@tmp = $key3 =~ m/[f](?=[4rv])/g)
           + scalar(@tmp = $key3 =~ m/[g](?=[5tb])/g)
           + scalar(@tmp = $key3 =~ m/[h](?=[6yn])/g)
           + scalar(@tmp = $key3 =~ m/[j](?=[7um])/g)
           + scalar(@tmp = $key3 =~ m/[k](?=[8i,])/g)
           + scalar(@tmp = $key3 =~ m/[l](?=[9o\.])/g)
           + scalar(@tmp = $key3 =~ m/[;](?=[0p\/\-\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key3 =~ m/[z](?=[1qa])/g)
           + scalar(@tmp = $key3 =~ m/[x](?=[2ws])/g)
           + scalar(@tmp = $key3 =~ m/[c](?=[3ed])/g)
           + scalar(@tmp = $key3 =~ m/[v](?=[4rf])/g)
           + scalar(@tmp = $key3 =~ m/[b](?=[5tg])/g)
           + scalar(@tmp = $key3 =~ m/[n](?=[6yh])/g)
           + scalar(@tmp = $key3 =~ m/[m](?=[7uj])/g)
           + scalar(@tmp = $key3 =~ m/[,](?=[8ik])/g)
           + scalar(@tmp = $key3 =~ m/[\.](?=[9ol])/g)
           + scalar(@tmp = $key3 =~ m/[\/](?=[0p;\-\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key3 =~ m/[\-](?=[0p;\/\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key3 =~ m/[\^](?=[0p;\/\-\\\@\[:\]_])/g)
           + scalar(@tmp = $key3 =~ m/[\\](?=[0p;\/\-\^\@\[:\]_])/g)
           + scalar(@tmp = $key3 =~ m/[\@](?=[0p;\/\-\^\\\[:\]_])/g)
           + scalar(@tmp = $key3 =~ m/[\[](?=[0p;\/\-\^\\\@:\]_])/g)
           + scalar(@tmp = $key3 =~ m/[:](?=[0p;\/\-\^\\\@\[\]_])/g)
           + scalar(@tmp = $key3 =~ m/[\]](?=[0p;\/\-\^\\\@\[:_])/g)
           + scalar(@tmp = $key3 =~ m/[_](?=[0p;\/\-\^\\\@\[:\]])/g);
    push(@$res, [scalar($count), $keycount - 1]);
    $count = scalar(@tmp = $key3 =~ m/[12345](?=[asdfg])/g)
           + scalar(@tmp = $key3 =~ m/[qwert](?=[zxcvb])/g)
           + scalar(@tmp = $key3 =~ m/[asdfg](?=[12345])/g)
           + scalar(@tmp = $key3 =~ m/[zxcvb](?=[qwert])/g)
           + scalar(@tmp = $key3 =~ m/[67890\-\^\\](?=[hjkl;:\]])/g)
           + scalar(@tmp = $key3 =~ m/[yuiop\@\[](?=[nm,\.\/_])/g)
           + scalar(@tmp = $key3 =~ m/[hjkl;:\]](?=[67890\-\^\\])/g)
           + scalar(@tmp = $key3 =~ m/[nm,\.\/_](?=[yuiop\@\[])/g);
    push(@$res, [scalar($count), $keycount - 1]);
    $count = scalar(@tmp = $key3 =~ m/[12345](?=[zxcvb])/g)
           + scalar(@tmp = $key3 =~ m/[zxcvb](?=[12345])/g)
           + scalar(@tmp = $key3 =~ m/[67890\-\^\\](?=[nm,\.\/])/g)
           + scalar(@tmp = $key3 =~ m/[nm,\.\/_](?=[67890\-\^\\])/g);
    push(@$res, [scalar($count), $keycount - 1]);
  } else {
    $count = scalar(@tmp = $key2 =~ m/(?: )[^ ]*[1][^ ]*(?= [^ ]*[qaz])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[2][^ ]*(?= [^ ]*[wsx])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[3][^ ]*(?= [^ ]*[edc])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[4][^ ]*(?= [^ ]*[rfv])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[5][^ ]*(?= [^ ]*[tgb])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[6][^ ]*(?= [^ ]*[yhn])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[7][^ ]*(?= [^ ]*[ujm])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[8][^ ]*(?= [^ ]*[ik,])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[9][^ ]*(?= [^ ]*[ol\.])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[0][^ ]*(?= [^ ]*[p;\/\-\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[q][^ ]*(?= [^ ]*[1az])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[w][^ ]*(?= [^ ]*[2sx])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[e][^ ]*(?= [^ ]*[3dc])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[r][^ ]*(?= [^ ]*[4fv])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[t][^ ]*(?= [^ ]*[5gb])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[y][^ ]*(?= [^ ]*[6hn])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[u][^ ]*(?= [^ ]*[7jm])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[i][^ ]*(?= [^ ]*[8k,])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[o][^ ]*(?= [^ ]*[9l\.])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[p][^ ]*(?= [^ ]*[0;\/\-\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[a][^ ]*(?= [^ ]*[1qz])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[s][^ ]*(?= [^ ]*[2wx])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[d][^ ]*(?= [^ ]*[3ec])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[f][^ ]*(?= [^ ]*[4rv])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[g][^ ]*(?= [^ ]*[5tb])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[h][^ ]*(?= [^ ]*[6yn])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[j][^ ]*(?= [^ ]*[7um])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[k][^ ]*(?= [^ ]*[8i,])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[l][^ ]*(?= [^ ]*[9o\.])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[;][^ ]*(?= [^ ]*[0p\/\-\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[z][^ ]*(?= [^ ]*[1qa])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[x][^ ]*(?= [^ ]*[2ws])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[c][^ ]*(?= [^ ]*[3ed])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[v][^ ]*(?= [^ ]*[4rf])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[b][^ ]*(?= [^ ]*[5tg])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[n][^ ]*(?= [^ ]*[6yh])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[m][^ ]*(?= [^ ]*[7uj])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[,][^ ]*(?= [^ ]*[8ik])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[\.][^ ]*(?= [^ ]*[9ol])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[\/][^ ]*(?= [^ ]*[0p;\-\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[\-][^ ]*(?= [^ ]*[0p;\/\^\\\@\[:\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[\^][^ ]*(?= [^ ]*[0p;\/\-\\\@\[:\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[\\][^ ]*(?= [^ ]*y[0p;\/\-\^\@\[:\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[\@][^ ]*(?= [^ ]*[0p;\/\-\^\\\[:\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[\[][^ ]*(?= [^ ]*[0p;\/\-\^\\\@:\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[:][^ ]*(?= [^ ]*[0p;\/\-\^\\\@\[\]_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[\]][^ ]*(?= [^ ]*[0p;\/\-\^\\\@\[:_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[_][^ ]*(?= [^ ]*[0p;\/\-\^\\\@\[:\]])/g);
    push(@$res, [scalar($count), $movecount - 1]);
    $count = scalar(@tmp = $key2 =~ m/(?: )[^ ]*[12345][^ ]*(?= [^ ]*[asdfg])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[qwert][^ ]*(?= [^ ]*[zxcvb])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[asdfg][^ ]*(?= [^ ]*[12345])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[zxcvb][^ ]*(?= [^ ]*[qwert])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[67890\-\^\\][^ ]*(?= [^ ]*[hjkl;:\]])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[yuiop\@\[][^ ]*(?= [^ ]*[nm,\.\/_])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[hjkl;:\]][^ ]*(?= [^ ]*[67890\-\^\\])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[nm,\.\/_][^ ]*(?= [^ ]*[yuiop\@\[])/g);
    push(@$res, [scalar($count), $movecount - 1]);
    $count = scalar(@tmp = $key2 =~ m/(?: )[^ ]*[12345][^ ]*(?= [^ ]*[zxcvb])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[zxcvb][^ ]*(?= [^ ]*[12345])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[67890\-\^\\][^ ]*(?= [^ ]*[nm,\.\/])/g)
           + scalar(@tmp = $key2 =~ m/(?: )[^ ]*[nm,\.\/_][^ ]*(?= [^ ]*[67890\-\^\\])/g);
    push(@$res, [scalar($count), $movecount - 1]);
  }
  
  $res;
  
}


sub parsekey {
  my $conf = shift;
  my $table = shift;
  my $teststr = shift;
  my $scorefunc = shift;
  
  my @pat = ([$teststr, '']);
  
  my @respat;
  
  while(1) {
    my $maxlen;
    my $minscore;
    foreach my $pat (@pat) {
      my $len = length($pat->[0]);
      my $score = &$scorefunc($pat->[1]);
      if(!defined($maxlen) or $maxlen < $len) {
        $maxlen = $len;
      }
      if(!defined($minscore) or $minscore > $score) {
        $minscore = $score;
      }
    }
    my @thispat;
    my @nextpat;
    foreach my $pat (@pat) {
      if(length($pat->[0]) == $maxlen) {
        next if(&$scorefunc($pat->[1]) > $minscore);
        next if(scalar(@thispat));
        push(@thispat, $pat);
      } else {
        push(@nextpat, $pat);
      }
    }
#    print "minscore: $minscore, maxlen: $maxlen, trypat: ".scalar(@thispat)."/" . scalar(@nextpat) . "\n";
    
    my @newpat;
    foreach my $pat (@thispat) {
      my $check;
      my $str = $pat->[0];
      my $strkana = &tensplit($str);
      foreach my $key (keys %$table) {
#        print "$key";
        if($str =~ s/^\Q$table->{$key}\E//) {
          my $newpat = [$str, $pat->[1] . ' ' . $key . '(' . $table->{$key} . ')'];
          if($str eq '') {
            push(@respat, $newpat->[1]);
          } else {
            push(@newpat, $newpat);
          }
          $check = 1;
#          print "$pat->[0]:$pat->[1] -> $str: $key, $table->{$key}\n";
          $str = $pat->[0];
        }
      }
      next if($check);
      foreach my $key (keys %$table) {
#        print "$key";
        if($strkana =~ s/^\Q$table->{$key}\E//) {
          my $newpat = [$strkana, $pat->[1] . ' ' . $key . '(' . $table->{$key} . ')'];
          if($strkana eq '') {
            push(@respat, $newpat->[1]);
          } else {
            push(@newpat, $newpat);
          }
#          print "$pat->[0]:$pat->[1] -> $strkana: $key, $table->{$key}\n";
          $strkana = &tensplit($str);
        }
      }
    }
    
#    print STDERR Data::Dumper->new([\@respat, \@thispat, \@newpat, \@nextpat])->Dump;
    if(scalar(@newpat)) {
      @pat = (@newpat, @nextpat);
    } else {
      if(scalar(@nextpat)) {
        @pat = (@nextpat);
      } else {
        last;
      }
    }
  }
  
  my $minpat;
  my $minscore;
  foreach my $pat (@respat) {
    my $score = &$scorefunc($pat);
    if(!defined($minscore) or $score < $minscore) {
      $minscore = $score;
      $minpat = $pat;
    }
  }
  
  $minpat =~ s/^ //;
  $minpat;
}


sub loadkeytable {
  my $file = shift;
  open(KEYFILE, $file) or die "Can't open $file";
  my $conf = {};
  my $table = {};
  my $prefix = '';
  
  while(<KEYFILE>) {
    next if(m/^'/);
    $_ =~ tr/\r\n//d;
    if(m/^(\w+)=(\w+)/) {
      $conf->{$1} = $2;
      next;
    }
    if(m/^=([^\t]*)/) {
      $prefix = $1;
      next;
    }
    if(m/^(.+)\t(.+)/) {
      my $key = $1;
      my $str = $2;
#      warn "$key/$str\n";
      while($key =~ s/^\{((?:$re_char)+?)\}|(.)//) {
        my $ch = defined($1) ? $1 : $2;
        if($str =~ s/^((?:$re_char)+?)(?:\$|$)//) {
          my $ch2 = $1;
          $ch2 = &hira2kata($ch2);
          $table->{$prefix.$ch} = $ch2;
#          warn "$prefix.$ch: $ch2\n";
        }
      }
      if($key ne '' or $str ne '') {
        warn "Parse error. ($file)[$_][$key/$str]\n";
      }
    }
  }
  ($conf, $table);
}



sub hira2kata {
  my $str = shift;
  $str =~ s/($re_char)/exists($HIRAKATA{$1}) ? $HIRAKATA{$1} : $1/eg;
  $str;
}

sub tensplit {
  my $str = shift;
  $str =~ s/($re_char)/exists($TENSPLIT{$1}) ? $TENSPLIT{$1} : $1/e;
  $str;
}

sub hue {
  my $h = shift;
  my $s = 1;
  my $v = 0.6;
  
  $h = ($h * 0.6 * 360) % 360;
  $h = $h / 60;
  my $i = int($h);
  my $f = $h - $i;
  
  if ($i == 0) { return sprintf('#%02x%02x%02x', $v*128, $v*(1 - $s*(1 - $f))*128, $v*(1-$s)*128); }
  if ($i == 1) { return sprintf('#%02x%02x%02x', $v*(1 - $s*$f)*128, $v*128, $v*(1-$s)*128); }
  if ($i == 2) { return sprintf('#%02x%02x%02x', $v*(1-$s)*128, $v*128, $v*(1 - $s*(1 - $f))*128); }
  if ($i == 3) { return sprintf('#%02x%02x%02x', $v*(1-$s)*128, $v*(1 - $s*$f)*128, $v*128); }
  if ($i == 4) { return sprintf('#%02x%02x%02x', $v*(1 - $s*(1 - $f))*128, $v*(1-$s)*128, $v*128); }
  if ($i == 5) { return sprintf('#%02x%02x%02x', $v*128, $v*(1-$s)*128, $v*(1 - $s*$f)*128); }
}


